
#ifndef MP_TCP_SERVER_H_
#define MP_TCP_SERVER_H_

#include "stdio_synchronizer.hpp"
#include "io_manager.hpp"
#include "tcp_client.hpp"
#include "tcp_socket.hpp"

#include <algorithm>
#include <atomic>
#include <condition_variable>
#include <functional>
#include <cstdint>
#include <list>
#include <memory>
#include <mutex>
#include <string>
#include <vector>
#include <queue>
#include <thread>
#include <unordered_map>

#ifdef _WIN32
#include <Winsock2.h>
#else
#include <sys/select.h>
#include <unistd.h>
#endif /* _WIN32 */

#include <cstdint>
#include <stdexcept>
#include <iostream>
#include <string>
#include <memory>
#include <mutex>

// Note:
//   If it is using auto type, it will be easier to read because #ifdef lines are reduced.
//   #ifdef was caused by the difference of socket file descriptor type on Windows and Ubuntu.

namespace multiplatform {

// multiplatform::tcp_server is the class providing TCP Server features.
// The tcp_server works entirely asynchronously, waiting for the io_manager to notify whenever a new client wished to connect.
class tcp_server {
public:

#ifdef _WIN32
  using socket_fd_t = SOCKET;
#else // _WIN32
  using socket_fd_t = int;
#endif // _WIN32

  tcp_server(void): m_io_manager(get_default_io_manager()), m_on_new_connection_callback(nullptr){
    if(m_debug){ _debug_print("create tcp_server"); }
  }
  ~tcp_server(void){
    if(m_debug){ _debug_print("destroy tcp_server"); }
    stop();
  }
  tcp_server(const tcp_server&) = delete;             // Removed copy constructor
  tcp_server& operator=(const tcp_server&) = delete;  // Removed assignment operator

  // Comparison operator
  bool operator==(const tcp_server& rhs) const{ return m_socket == rhs.m_socket; }
  bool operator!=(const tcp_server& rhs) const{ return m_socket != rhs.m_socket; }

  // callback called whenever a new client is connecting to the server
  //
  // Takes as parameter a shared pointer to the tcp_client that wishes to connect
  // Returning true means connection is handled by tcp_client wrapper and nothing will be done by tcp_server. Returning false means connection is handled by tcp_server, will be stored in an internal list and tcp_client disconection_handler overriden.
  using on_new_connection_callback_t = std::function<bool(const std::shared_ptr<tcp_client>&)>;

  // Start the tcp_server at the given host and port.
  //
  // \param host hostname to be connected to
  // \param port port to be connected to
  // \param callback callback to be called on new connections (may be null, connections are then handled automatically by the tcp_server object)
  void start(const std::string& host, std::uint32_t port, const on_new_connection_callback_t& callback = nullptr){
    if(is_running()){
      _error_throw("tcp_server is already running");
    }
    m_socket.bind(host, port);
    m_socket.listen(1024); // connection queue size
    m_io_manager->track(m_socket);
    m_io_manager->set_rd_callback(m_socket, std::bind(&tcp_server::on_read_available, this, std::placeholders::_1));
    m_on_new_connection_callback = callback;
    m_is_running = true;
    if(m_debug){ _info_print("tcp_server running"); }
  }

  // Disconnect the tcp_server if it was currently running.
  // \param wait_for_removal When sets to true, disconnect blocks until the underlying TCP server has been effectively removed from the io_manager and that all the underlying callbacks have completed.
  // \param recursive_wait_for_removal When sets to true and wait_for_removal is also set to true, blocks until all the underlying TCP client connected to the TCP server have been effectively removed from the io_manager and that all the underlying callbacks have completed.
  void stop(bool wait_for_removal = false, bool recursive_wait_for_removal = true){
    if(!is_running()){ return; }
    m_is_running = false;
    m_io_manager->untrack(m_socket);
    if(wait_for_removal){ m_io_manager->wait_for_removal(m_socket); }
    m_socket.close();
    std::lock_guard<std::mutex> clients_lock(m_clients_mutex);
    for(auto& client : m_clients){
      client->disconnect(recursive_wait_for_removal && wait_for_removal);
    }
    m_clients.clear();
    if(m_debug){ _info_print("tcp_server stopped"); }
  }

  // \return whether the server is currently running or not
  bool is_running(void) const{ return m_is_running; }

  // \return the tcp_socket associated to the server. (non-const/const version)
  multiplatform::tcp_socket& get_socket(void){ return m_socket; }
  const multiplatform::tcp_socket& get_socket(void) const{ return m_socket; }

  // \return io service monitoring this tcp connection
  const std::shared_ptr<multiplatform::io_manager>& get_io_manager(void) const{ return m_io_manager; }

  // \return the list of tcp_client connected to the server.
  const std::list<std::shared_ptr<multiplatform::tcp_client>>& get_clients(void) const{ return m_clients; }

private:
  // io manager read callback
  //
  // \param fd socket that triggered the read callback
  void on_read_available(socket_fd_t fd){
    try{
      if(m_debug){ _info_print("tcp_server received new connection"); }
      auto client = std::make_shared<tcp_client>(m_socket.accept());
      if(!m_on_new_connection_callback || !m_on_new_connection_callback(client)){
        if(m_debug){ _info_print("connection handling delegated to tcp_server"); }
        client->set_on_disconnection_handler(std::bind(&tcp_server::on_client_disconnected, this, client));
        m_clients.push_back(client);
      }
      else{
        if(m_debug){ _info_print("connection handled by tcp_server wrapper"); }
      }
    }
    catch(const std::runtime_error&){
      _warning_print("accept operation failure");
      stop();
    }
  }

  // client disconnected
  // called whenever a client disconnected from the tcp_server
  //
  // \param client disconnected client
  void on_client_disconnected(const std::shared_ptr<tcp_client>& client){
    // If we are not running the server
    // Then it means that this function is called by tcp_client::disconnect() at the destruction of all clients
    if(!is_running()){ return; }
    if(m_debug){ _debug_print("handle server's client disconnection"); }
    std::lock_guard<std::mutex> clients_lock(m_clients_mutex);
    auto it = std::find(m_clients.begin(), m_clients.end(), client);
    if(it != m_clients.end()){ m_clients.erase(it); }
  }

  // store io_manager
  // prevent deletion of io_manager before the tcp_server itself
  std::shared_ptr<io_manager> m_io_manager;

  multiplatform::tcp_socket m_socket; // Server socket

  // whether the server is currently running or not
  std::atomic<bool> m_is_running = ATOMIC_VAR_INIT(false);

  std::list<std::shared_ptr<multiplatform::tcp_client>> m_clients;  // clients
  std::mutex m_clients_mutex;                                       // Clients thread safety
  //bool m_debug;                                                     // Print debug messages for this class.

  // on new connection callback
  on_new_connection_callback_t m_on_new_connection_callback;

  std::atomic<bool> m_debug = ATOMIC_VAR_INIT(false);
};

} // namespace multiplatform

#endif // MP_TCP_SERVER_H_

