
#ifndef MP_THREAD_POOL_H_
#define MP_THREAD_POOL_H_

#include "stdio_synchronizer.hpp"
#include <atomic>
#include <condition_variable>
#include <functional>
#include <list>
#include <mutex>
#include <queue>
#include <thread>
#include <vector>

#include <cstdint>
#include <stdexcept>
#include <iostream>
#include <string>
#include <memory>
#include <mutex>

namespace multiplatform {

// basic thread pool used to push async tasks from the io_service
// Notes: c++11 better implements refere to following, but following implementation needs '#include <future>'.
//   http://progsch.net/wordpress/?p=81
//   https://github.com/progschj/ThreadPool
class thread_pool {
public:
  explicit thread_pool(std::size_t counts){
    if(m_debug){ _debug_print("creating thread_pool"); }
    set_max_threads(counts);
    if(m_debug){ _debug_print("thread_pool created"); }
  }
  ~thread_pool(void){
    if(m_debug){ _debug_print("discarding thread_pool"); }
    stop();
    if(m_debug){ _debug_print("thread_pool discarded"); }
  }
  thread_pool(const thread_pool&) = delete;            // removed copy constructor
  thread_pool& operator=(const thread_pool&) = delete; // removed assignment operator

  // Add tasks to thread pool.
  // task is enqueued, and will be executed whenever the next timing comes.
  // 1. All previously executed tasked have been executed.
  // 2. Tasks are currently being executed.
  void add_task(const std::function<void()>& task){
    if(m_debug){ _debug_print("add a task to thread_pool"); }
    std::lock_guard<std::mutex> task_lock(m_tasks_mutex);
    m_tasks.push(task);
    m_tasks_condition.notify_one();
    if(m_debug){ _debug_print("task has been added to thread_pool"); }
  }

  // The add_task() wrapper to support the following syntax.
  // thread_pool tasks; tasks << task1 << task2 << task3;
  thread_pool& operator<<(const std::function<void()>& task){
    add_task(task);
    return *this;
  }

  // Stop the thread pool and wait for workers completion.
  // If some tasks are pending they will not be executed.
  void stop(void){
    if(!is_running()){ return; }
    m_should_stop = true;
    m_tasks_condition.notify_all();
    for(auto& worker : m_workers){ worker.join(); }
    m_workers.clear();
    if(m_debug){ _debug_print("thread_pool stopped"); }
  }

  bool is_running(void) const{ return !m_should_stop; }

  // This function returns immediately, but changes may be applied in the background.
  // Changing the number of workers does not affect the task to be executed or the currently running task.
  void set_max_threads(std::size_t counts){
    m_max_threads = counts;
    while(m_running_threads < m_max_threads){
      // If increasing the number of threads spawn a new thread although it may take some time to start.
      m_running_threads++;
      m_workers.push_back(std::thread(std::bind(&thread_pool::run, this)));
    }
    if(m_running_threads > m_max_threads){
      // If decrease the number of threads, start the thread in the background,
      // stop the thread until the appropriate amount of thread is obtained,
      // and so that the thread does not stop in the middle of the task.
      m_tasks_condition.notify_all();
    }
  }

private:
  // Main loop of worker
  void run(void){
    if(m_debug){ _debug_print("start run() worker"); }
    while(true){
      std::pair<bool, std::function<void()> > ret = fetch_task_or_stop();
      if(ret.first){ break; } // If thread has been requested to stop (ret.first==true), stop it here.
      if(ret.second){
        if(m_debug){ _debug_print("execute task"); }
        try{ ret.second(); } // Try to execute task. (ret.second is task)
        catch(const std::exception&){
          _warning_print("uncatched exception propagated up to the thread_pool.");
        }
        if(m_debug){ _debug_print("task completed"); }
      }
    }
    if(m_debug){ _debug_print("stop run() worker"); }
  }

  // Fetch the first element (task) in the queue or wait if there are no available tasks.
  std::pair<bool, std::function<void()> > fetch_task_or_stop(void){
    if(m_debug){ _debug_print("waiting to fetch a task"); }
    std::unique_lock<std::mutex> task_lock(m_tasks_mutex);
    m_tasks_condition.wait(task_lock, [&]{ return should_stop() || !m_tasks.empty(); });
    if(should_stop()){
      if(m_debug){ _debug_print("should be stop a thread"); }
      m_running_threads--;
      return {true, nullptr};
    }
    std::function<void()> task = std::move(m_tasks.front());
    m_tasks.pop();
    if(m_debug){ _debug_print("fetched a task"); }
    return {false, task};
  }

  bool should_stop(void) const{ return m_should_stop || m_running_threads > m_max_threads; }

  std::atomic<bool> m_debug = ATOMIC_VAR_INIT(false);              // Print debug messages for this class.
  std::atomic<bool> m_should_stop = ATOMIC_VAR_INIT(false);        // Whether thread_pool should be stop or not.
  std::atomic<std::size_t> m_max_threads = ATOMIC_VAR_INIT(0);     // Number of allowed threads.
  std::atomic<std::size_t> m_running_threads = ATOMIC_VAR_INIT(0); // Number of running threads.
  std::list<std::thread> m_workers;                                // Worker threads.
  std::queue<std::function<void()> > m_tasks;                      // Queue of the task.
  std::mutex m_tasks_mutex;                                        // Mutex for the thread safe of the tasks.
  std::mutex m_cerr_mutex;                                         // Prevent corruption of error messages.
  std::condition_variable m_tasks_condition;                       // Tasks condition to synchronizing task changes.
};

} // namespace multiplatform

#endif // MP_THREAD_POOL_H_

